import asyncio
import typing as t
import json
import abc
import warnings
import socket
import io
import os
import threading
import types
import gzip
import time
import logging
import datetime
import itertools
from os import path
from contextlib import AbstractAsyncContextManager
import urllib.parse as urlparse

true_strings: t.Set[str] = {"true", "1", "yes", "on"}

log = logging.getLogger(__name__)

T = t.TypeVar("T")

URL = t.Union[urlparse.ParseResult, str]

class GPSProvider(abc.ABC):
    async def __aenter__(self: T) -> T:
        return self
    
    async def __aexit__(self, exc_type: t.Optional[t.Type[BaseException]], exc_value: t.Optional[BaseException], traceback: t.Optional[types.TracebackType]) -> None:
        await self.close()

    @property
    @abc.abstractmethod
    def url(self) -> urlparse.ParseResult: ... # pragma: no cover

    @property
    @abc.abstractmethod
    def query(self) -> t.Dict[str, t.List[str]]: ... # pragma: no cover

    
    @abc.abstractmethod
    async def connect(self) -> None: ... # pragma: no cover
        
    @abc.abstractmethod
    async def close(self) -> None: ... # pragma: no cover

    @abc.abstractmethod
    async def next(self) -> t.Dict[str, t.Any]: ... # pragma: no cover

    @abc.abstractmethod
    async def watch(self, enable: bool = True) -> None: ... # pragma: no cover

    def __aiter__(self) -> t.AsyncIterator[t.Dict[str, t.Any]]:
        return self

    async def __anext__(self) -> t.Dict[str, t.Any]:
        try:
            return await self.next()
        except EOFError:
            raise StopAsyncIteration()

class RootProvider(GPSProvider, abc.ABC):
    url: urlparse.ParseResult = t.cast(urlparse.ParseResult, None) # this is here to satisfy the abc.
    query: t.Dict[str, t.List[str]] = t.cast(t.Dict[str, t.List[str]], None) # this is here to satisfy the abc.
    export_comments: bool

    def __init__(self, url: URL, export_comments: bool = False) -> None:
        if isinstance(url, str):
            url = urlparse.urlparse(url)
        self.url = url
        self.query = urlparse.parse_qs(url.query)
        self.export_comments = export_comments
    
    @abc.abstractmethod
    async def _read(self) -> bytes: ... # pragma: no cover

    async def next(self) -> t.Dict[str, t.Any]:
        while True:
            line = await self._read()

            if len(line) == 0: # EOF
                await self.close()
                raise EOFError()
            
            if line in (b"\n", b"\r\n"):
                continue
            
            if line[0:1] == b"#":
                if self.export_comments:
                    return { 'class':'COMMENT', 'comment': line.strip() }
                else:
                    continue # pragma: no cover

            js = json.loads(line)
            return t.cast(t.Dict[str, t.Any], js)
    

class GPSDProvider(RootProvider):
    reader: t.Optional[asyncio.StreamReader]
    writer: t.Optional[asyncio.StreamWriter]
    
    def __init__(self, url: URL = "tcp://localhost", export_comments: bool = False) -> None:
        super().__init__(url, export_comments)
        self.reader = None
        self.writer = None

    async def connect(self) -> None:
        if self.writer is not None:
            raise ValueError("called connect on connected GPSDProvider.")
        if self.url.scheme == "tcp":
            self.reader, self.writer = await asyncio.open_connection(
                host=self.url.hostname,
                port=self.url.port if self.url.port is not None else 2947)
        elif self.url.scheme == "unix":
            self.reader, self.writer = await asyncio.open_unix_connection(path=self.url.path)
        elif self.url.scheme in ("fd", "socket"):
            self.reader, self.writer = await asyncio.open_connection(sock=socket.socket(fileno=os.dup(int(self.url.netloc))))
        else:
            raise ValueError(f"invalid scheme {self.url.scheme} for GPSDProvider.")

    async def _read(self) -> bytes:
        if self.reader is None:
            raise ValueError("called next on unconnected GPSDProvider.")
        return await self.reader.readline()

    async def close(self) -> None:
        if self.writer is not None:
            self.writer.close()
            await self.writer.wait_closed()
            self.writer = None
            self.reader = None
    
    async def watch(self, enable: bool = True) -> None:
        if self.writer is None:
            raise ValueError("called watch on not connected GPSDProvider")
        
        self.writer.write(b"?WATCH="+json.dumps({"enable": enable, "json": True}, ensure_ascii=True).encode("ascii"))
        await self.writer.drain()

class FileProvider(RootProvider):
    file: t.Optional[t.Union[t.BinaryIO, gzip.GzipFile]]
    enabled: asyncio.Event

    def __init__(self, url: URL, export_comments: bool = False):
        super().__init__(url, export_comments)
        self.file = None
        self.enabled = asyncio.Event()
    
    async def connect(self) -> None:
        if self.file is not None:
            raise ValueError("called connect on connected FileProvider.")
        if self.url.scheme in ("file", ""):
            if self.url.path.endswith(".gz"):
                self.file = gzip.open(self.url.path, mode="rb")
            else:
                self.file = open(self.url.path, "rb")
        elif self.url.scheme == "fd":
            self.file = open(int(self.url.netloc), "rb", closefd=False)
        else:
            raise ValueError(f"invalid scheme {self.url.scheme} for FileProvider.")

    async def _read(self) -> bytes:
        if self.file is None:
            raise ValueError("next called on unconnected FileProvider")
        
        return self.file.readline()

    async def next(self) -> t.Dict[str, t.Any]:
        ret = await super().next()
        # wait for activation if event comes from a device or is a device activation notification.
        if "device" in ret or (ret["class"] == "DEVICE" and "activated" in ret and ret["activated"] != 0):
            await self.enabled.wait()
        return ret

    async def close(self) -> None:
        if self.file is not None:
            self.enabled.clear()
            self.file.close()
            self.file = None
        
    async def watch(self, enable: bool = True) -> None:
        if self.file is None:
            raise ValueError("watch called on unconnected FileProvider")
        
        if enable:
            self.enabled.set()
        else:
            self.enabled.clear()

class DelegateProvider(GPSProvider):
    delegate: GPSProvider
    
    def __init__(self, delegate: GPSProvider):
        self.delegate = delegate

    @property
    def url(self) -> urlparse.ParseResult:
        return self.delegate.url

    @property
    def query(self) -> t.Dict[str, t.List[str]]:
        return self.delegate.query

    
    async def connect(self) -> None:
        await self.delegate.connect()

    async def next(self) -> t.Dict[str, t.Any]:
        return await self.delegate.next()

    async def close(self) -> None:
        await self.delegate.close()

    async def watch(self, enable: bool = True) -> None:
        await self.delegate.watch(enable)

class Throttle(DelegateProvider):
    last_clockval: t.Optional[float]
    interval: float
    
    def __init__(self, delegate: GPSProvider):
        super().__init__(delegate)
        
        self.interval = float(self.query["interval"][-1])
        
        self.last_clockval = None

    async def connect(self) -> None:
        await super().connect()

        self.last_clockval = None

    async def next(self) -> t.Dict[str, t.Any]:
        obj = await super().next()
        
        if obj["class"] == "TPV":
            
            clockval = time.monotonic()
            
            if self.last_clockval is not None:
                elapsed = clockval-self.last_clockval
                wait_time = self.interval - elapsed
                
                if wait_time > 0:
                    
                    await asyncio.sleep(wait_time)
                    
                    self.last_clockval += self.interval
                                
                else:
                    log.warning("Data from source [%s] is arriving too slow for interval [%s]"%(str(self.delegate),self.interval))
                    self.last_clockval = clockval
                        
            else:
                self.last_clockval = clockval

        return obj

class FakeTimestamps(DelegateProvider):
    async def next(self) -> t.Dict[str, t.Any]:
        ret = await super().next()

        if "time" in ret:
            ret["time"] = datetime.datetime.now(datetime.timezone.utc).isoformat()

        return ret

class ShiftTimestamps(DelegateProvider):
    start_timestamp: datetime.datetime
    offset: t.Optional[datetime.timedelta]
    def __init__(self, delegate: GPSProvider) -> None:
        super().__init__(delegate)
        
        self.start_timestamp = datetime.datetime.fromisoformat(self.query["start_timestamp"][-1])
        self.offset = None

    async def connect(self) -> None:
        await super().connect()
        self.offset = None
        
    async def next(self) -> t.Dict[str, t.Any]:
        ret = await super().next()
        
        if "time" in ret:
            time = datetime.datetime.fromisoformat(ret["time"])
            if self.offset is None:
                self.offset = self.start_timestamp - time
                ret["time"] = self.start_timestamp.isoformat()
            else:
                ret["time"] = (time + self.offset).isoformat()

        return ret

class Loop(GPSProvider):
    delegate: t.Optional[GPSProvider]
    # workaround for https://github.com/python/mypy/issues/708
    delegate_factory: t.Union[t.Callable[[URL], GPSProvider], t.Callable[[URL], GPSProvider]]
    delegate_url_template: urlparse.ParseResult
    urls_it: t.Optional[t.Iterator[urlparse.ParseResult]]
    url: urlparse.ParseResult = ... # type: ignore # this is here to satisfy the abc.
    query: t.Dict[str, t.List[str]] = ... # type: ignore # this is here to satisfy the abc.
    enabled: bool
    endless: bool

    def __init__(self, url: URL, delegate_factory: t.Callable[[URL], GPSProvider]):
        if isinstance(url, str):
            url = urlparse.urlparse(url)
        self.url = url
        self.query = urlparse.parse_qs(url.query)
        endless_s = self.query.get("endless")
        self.endless = endless_s is not None and endless_s[-1] in true_strings
        self.delegate = None
        self.delegate_factory = delegate_factory
        template_query = self.query.copy()
        if endless_s is not None:
            del template_query["endless"]
        self.delegate_url_template = url._replace(query=urlparse.urlencode(template_query, doseq=True))
        self.enabled = False
        self.urls_it = None
    
    async def _connect_next(self) -> None:
        assert self.urls_it is not None
        
        if self.delegate is not None:
            await self.delegate.close()

        try:
            self.delegate = self.delegate_factory(next(self.urls_it))
        except StopIteration:
            self.delegate = None
            await self.close()
            raise EOFError()
        
        await self.delegate.connect()
        if self.enabled:
            await self.delegate.watch()
    
    async def connect(self) -> None:
        if self.urls_it is not None:
            raise ValueError("called connect on connected Loop.")
        if self.url.scheme in ("", "file") and path.isdir(self.url.path):
            with os.scandir(self.url.path) as it:
                urls = [self.delegate_url_template._replace(path=i.path) for i in it if i.name.endswith(".gpstrack") or i.name.endswith(".gpstrack.gz") and i.is_file()]
            urls.sort()
        else:
            urls = [ self.url ]
        
        self.urls_it = itertools.cycle(urls) if self.endless else iter(urls)

        try:
            await self._connect_next()
        except EOFError:
            raise ValueError("directory is empty.")

    async def close(self) -> None:
        self.urls_it = None
        self.enabled = False
        if self.delegate is not None:
            await self.delegate.close()
            self.delegate = None

    async def watch(self, enable: bool = True) -> None:
        if self.delegate is None:
            raise ValueError("watch called on unconnected Loop.")

        if self.enabled != enable:
            await self.delegate.watch(enable)
        self.enabled = enable
        
    async def next(self) -> t.Dict[str, t.Any]:
        if self.delegate is None:
            raise ValueError("next called on unconnected Loop.")
        
        while True:
            try:
                return await self.delegate.next()
            except EOFError:
                await self._connect_next()
            
        

def provider_factory(url: URL) -> GPSProvider:
    if isinstance(url, str):
        url = urlparse.urlparse(url)

    query = urlparse.parse_qs(url.query)
    
    ret: GPSProvider

    if "endless" in query:
        ret = Loop(url, provider_factory)
    else:
        if url.scheme in ("", "file"):
            if path.isdir(url.path):
                ret = Loop(url, provider_factory)
            else:
                ret = FileProvider(url)
        elif url.scheme == "fd":
            ret = FileProvider(url)
        else:
            ret = GPSDProvider(url)

    
    if "interval" in query:
        ret = Throttle(ret)

    fake_timestamps = query.get("fake_timestamps")
    if fake_timestamps is not None and fake_timestamps[-1].lower() in true_strings:
        ret = FakeTimestamps(ret)

    if "start_timestamp" in query:
        ret = ShiftTimestamps(ret)

    return ret
    
async def main(args: t.List[str]) -> None:
    async with provider_factory(args[0]) as gps:
        await gps.connect()
        await gps.watch()
        async for i in gps:
            print(i)

if __name__=="__main__":
    logging.basicConfig(level=logging.INFO)
    from sys import argv
    asyncio.run(main(argv[1:]))
