'''

  Support for faking timestamps ini GPSD JSON packages.

'''

from clazzes.gpsclient.helpers import DelegatorSupport
import time
import logging
from urllib.parse import urlparse, parse_qs
from clazzes.tools.lang import str2bool
import datetime
from clazzes.tools import datetools

log = logging.getLogger(__name__)

class FakeTimestamps(DelegatorSupport):
    
    def __init__(self,delegate,fake_timestamps = True):
        '''
        :param delegate: The delegate to fetch GPSD JSON packages from.
        :param fake_timestamps: Whether to apply timestamp faking to TPV and SKY records.
        '''    
        super().__init__(delegate)
        self.fake_timestamps = fake_timestamps
        self.ts_offset = None
    
    def connect(self, addr="tcp://localhost?fake_timestamps=true"):
        
        if isinstance(addr,str):
            addr = urlparse(addr)

        fake_timestamps_s = parse_qs(addr.query).get('fake_timestamps')
        
        if fake_timestamps_s is not None:
            self.fake_timestamps = str2bool(fake_timestamps_s[0])
            
        super().connect(addr=addr)

    def fake_timestamp(self,obj):
        time = obj.get("time")
        
        # only fake timestamps, which are present in the source
        # object.
        # (We never observed timestamps in SKY records, although GPSD
        #  JSON spec claims, that they might be present...)
        
        if time is not None:
            obj["time"]=datetools.format_datetime(datetime.datetime.now(tz=datetime.timezone.utc))
        
    def next(self, timeout=0):
        '''
        Fetch the next package from the delegate and fake timestamps of
        TPV and SKY records, if ``fake_timestamps`` was configured to true.
        
        :param timeout: The timeout in seconds to wait for a package on the
            the delegate.
        '''
        obj = super().next(timeout=timeout)
        
        if obj is not None and self.fake_timestamps:
            
            cls = obj.get("class")
            
            if cls == "TPV" or cls == "SKY":
                self.fake_timestamp(obj)
                    
        return obj
    
    
    def __str__(self):
        
        base = str(self.delegate)
        return "%s%sfake_timestamps=%s"%(base,'&' if '?' in base else '?',"true" if self.fake_timestamps else "false")

class ShiftTimestamps(DelegatorSupport):
    
    def __init__(self,delegate,start_timestamp = None):
        '''
        :param delegate: The delegate to fetch GPSD JSON packages from.
        :param start_timestamp: A start timestamp to shift TPV and SKY record times to.
                 The first timestamp is mapped onto this start timestamp and all subsequent
                timestamps are then shifted by the time interval given by the translation
                of the first timestamp to the given one.
        ''' 
        super().__init__(delegate)
        
        if isinstance(start_timestamp,str):
            self.start_timestamp = datetools.parse_datetime(start_timestamp,tzinfo=datetime.timezone.utc)
        else:
            self.start_timestamp = start_timestamp
            
        self.ts_offset = None
    
    def connect(self, addr="tcp://localhost"):
        
        if isinstance(addr,str):
            addr = urlparse(addr)

        start_timestamp_s = parse_qs(addr.query).get('start_timestamp')
        
        if start_timestamp_s is not None:
            self.start_timestamp = datetools.parse_datetime(start_timestamp_s[0],tzinfo=datetime.timezone.utc)
        
        self.ts_offset = None

        super().connect(addr=addr)
    
    def next(self, timeout=0):
        '''
        Fetch the next package from the delegate and translate timestamps of
        TPV and SKY records, if ``start_timestamp`` was configured to an alternate start timestamp.
        
        :param timeout: The timeout in seconds to wait for a package on the
            the delegate.
        '''
        obj = super().next(timeout=timeout)
        
        if obj is not None and self.start_timestamp is not None:
            
            cls = obj.get("class")
            
            if cls == "TPV" or cls == "SKY":
                
                time = obj.get("time")
                
                # only shift timestamps, which are present in the source
                # object.
                # (We never observed timestamps in SKY records, although GPSD
                #  JSON spec claims, that they might be present...)
                 
                if time is not None:
                    
                    dt = datetools.parse_datetime(time,tzinfo=datetime.timezone.utc)
                    
                    if self.ts_offset is None:
                        self.ts_offset = self.start_timestamp - dt
                        obj["time"]=datetools.format_datetime(self.start_timestamp)
                        
                        log.info("Mapping first timestamp [%s] of [%s] to [%s], time difference is [%s]."%
                                 (time,str(self.delegate),obj["time"],self.ts_offset))
                    else:
                        obj["time"]=datetools.format_datetime(dt+self.ts_offset)

        return obj
    
    def __str__(self):
        
        base = str(self.delegate)
        if self.start_timestamp is None:
            return base
        
        return "%s%sstart_timestamp=%s"%(base,'&' if '?' in base else '?',datetools.format_datetime(self.start_timestamp))
