#!/usr/bin/python3
'''
Main program for displaying GPSD responses and storing log files.
'''

import sys
import getopt

import logging
import signal
import curses
import datetime
import time
import re
import json

from urllib.parse import urlparse

from clazzes.tools import datetools

from clazzes.gpsclient.gpslog import LogWriter
from clazzes.gpsclient.logserver import run_server
from clazzes.gpsclient.factory import new_client
from collections import defaultdict
from heapq import nlargest

log = logging.getLogger(__name__)

interrupted = False
interruptor = None

def sighandler(signum,fram):
    global interrupted
    global interruptor
    log.info("Interrupted by signal [%d]"%signum)
    interrupted = True
    if interruptor:
        interruptor()

def _register_interruptor(_interruptor):
    global interruptor
    interruptor = _interruptor
    
def _open_client(url,logfile,export_comments=False):
    
    if logfile is None:
        return new_client(url,export_comments=export_comments)
    else:
        ret = LogWriter(delegate=new_client(url,export_comments=export_comments))
        log.info("Opening log file [%s] for writing."%logfile)
        ret.open(logfile)
        return ret

def run_gps_client(url,logfile,quiet=False,enable=True,devicepath=None,initial_timeout=10,timeout=10,print_infos=False):
    """
    Connect to a GPSD server and receive positions to be stored to the database. 
    """
    
    with _open_client(url,logfile,export_comments=print_infos) as client:
    
        client.connect(url)
        client.watch(enable,devicepath)

        real_timeout = initial_timeout

        if print_infos:
            record_count = defaultdict(int)
        else:
            record_count = None
            
        first_timestamp = None
        last_timestamp = None

        while not interrupted:
    
            try:
                data = client.next(timeout=real_timeout)
                
                if not data:
                    log.info("GPSD client got no response for %s seconds, leaving main loop."%timeout)
                    break
                else:
                    if not quiet:
                        log.info("Got response [%s] from GPSD."%json.dumps(data,sort_keys=True))

                    cls = data.get('class')

                    if record_count is not None and cls is not None and cls != "COMMENT":
                        record_count[cls] += 1

                    if 'COMMENT' == cls:
                        if first_timestamp is None:
                            comment = data.get("comment")
                            if comment is not None:
                                log.info(comment)

                    elif 'TPV' == cls:
                        
                        time_s = data.get('time')
                        
                        if time_s is not None:
                            
                            last_timestamp = time_s
                            if first_timestamp is None:
                                first_timestamp = time_s
                                log.info("Received first timestamp [%s]."%first_timestamp)

                        if timeout != real_timeout:
                            real_timeout = timeout

            except InterruptedError:
                log.info("Wait for GPSD response has been interrupted.")

        log.info("Received timestamps from [%s] to [%s]."%(first_timestamp,last_timestamp))
        
        if record_count is not None:
            log.info("      Record | count")
            log.info("  -----------+--------")

            for k in sorted(record_count.keys()):

                log.info("  %10s | %d"%(k,record_count[k]))

def _add_value_str(win,y,x,lbl,fmt,data,key):
    
    v = data.get(key)
    
    if v is None:
        w = int(re.search('%([0-9]+)',"%20s").group(1))
        if lbl:
            win.addstr(y,x,lbl+": -" + " " *(w-1))
        else:
            win.addstr(y,x,"-" + " " *(w-1))
    else:
        if lbl:
            win.addstr(y,x,(lbl+": "+fmt)%v)
        else:
            win.addstr(y,x,fmt%v)

def _add_msg(win,msg,y=2,x=5):
    h,w = win.getmaxyx()  # @UnusedVariable
    win.addstr(y,x,msg.ljust(w-x)[0:w-x])
    
def _set_cursor(win):
    h,w = win.getmaxyx()  # @UnusedVariable
    win.move(h-15,0)

def run_gps_client_curses(stdscr,url,logfile,enable=True,devicepath=None,initial_timeout=10,timeout=10):
    """
    Connect to a GPSD server and receive positions to be stored to the database. 
    """
    
    curses.curs_set(0)
    curses.nl()

    tpv_win = curses.newwin(18,31,5,5)
    sky_win = curses.newwin(18,38,5,38)
    sat_win = sky_win.derwin(15,16,2,20)

    while not interrupted:
    
        with _open_client(url,logfile) as client:
        
            if interrupted:
                break
        
            stdscr.clear()
            _add_msg(stdscr,"Connecting to [%s]..."%url.geturl())
            _set_cursor(stdscr)
            stdscr.refresh()
            
            try:
                client.connect(url)
                client.watch(enable,devicepath)
            except IOError:
                if interrupted:
                    _add_msg(stdscr,"Connection to [%s] failed, leaving."%url.geturl())
                    _set_cursor(stdscr)
                    stdscr.refresh()
                    return
                else:
                    _add_msg(stdscr,"Connection to [%s] failed, reconnecting..."%url.geturl())
                    _set_cursor(stdscr)
                    stdscr.refresh()
                    time.sleep(1)
                    continue
    
            stdscr.clear()
            _add_msg(stdscr,"Successfully connected to [%s]."%url.geturl())
            _set_cursor(stdscr)
            stdscr.refresh()
            
            tpv_win.clear()
            tpv_win.box()
            tpv_win.addstr(0,2,"TPV")
            tpv_win.refresh()
            
            sky_win.clear()
            sky_win.box()
            sky_win.addstr(0,2,"SKY")
            sky_win.refresh()

            real_timeout = initial_timeout

            while not interrupted:
    
                try:
                    data = client.next(timeout=real_timeout)
                    
                    if not data:
                        if interrupted:
                            _add_msg(stdscr,"Lost connection to [%s], leaving."%url.geturl())
                            _set_cursor(stdscr)
                            stdscr.refresh()
                            return
                        else:
                            _add_msg(stdscr,"Lost connection to [%s], reconnecting..."%url.geturl())
                            _set_cursor(stdscr)
                            stdscr.refresh()
                        
                        log.info("GPSD client got no response for %s seconds, leaving main loop."%timeout)
                        time.sleep(1)
                        break
                    else:
                        cls = data["class"]
                        
                        if timeout != real_timeout and 'TPV' == cls:
                            real_timeout = timeout

                        if cls == "TPV":
                            tmp = { "now": datetools.format_datetime(datetime.datetime.utcnow()) + "Z" }
                            
                            _add_value_str(tpv_win,1,1,"now","%24s",tmp,"now")
                            _add_value_str(tpv_win,2,1,"tim","%24s",data,"time")
                            _add_value_str(tpv_win,3,1,"mod","%5d",data,"mode")
                            _add_value_str(tpv_win,4,1,"lon","%12.6f",data,"lon")
                            _add_value_str(tpv_win,5,1,"lat","%12.6f",data,"lat")
                            _add_value_str(tpv_win,6,1,"alt","%12.6f",data,"alt")
                            _add_value_str(tpv_win,7,1,"spd","%7.1f",data,"speed")
                            _add_value_str(tpv_win,8,1,"cog","%7.1f",data,"track")
                            _add_value_str(tpv_win,9,1,"cli","%7.1f",data,"climb")
                            _add_value_str(tpv_win,10,1,"ept","%12.6f",data,"ept")
                            _add_value_str(tpv_win,11,1,"epx","%12.6f",data,"epx")
                            _add_value_str(tpv_win,12,1,"epy","%12.6f",data,"epy")
                            _add_value_str(tpv_win,13,1,"epv","%12.6f",data,"epv")
                            _add_value_str(tpv_win,14,1,"eps","%12.6f",data,"eps")
                            _add_value_str(tpv_win,15,1,"epd","%12.6f",data,"epd")
                            _add_value_str(tpv_win,16,1,"epc","%12.6f",data,"epc")
                            tpv_win.refresh()
    
                        if cls == "SKY":
                            _add_value_str(sky_win,1,1,"time","%24s",data,"time")
                            _add_value_str(sky_win,2,1,"tdop","%12.6f",data,"tdop")
                            _add_value_str(sky_win,3,1,"hdop","%12.6f",data,"hdop")
                            _add_value_str(sky_win,4,1,"xdop","%12.6f",data,"xdop")
                            _add_value_str(sky_win,5,1,"ydop","%12.6f",data,"ydop")
                            _add_value_str(sky_win,6,1,"vdop","%12.6f",data,"vdop")
                            _add_value_str(sky_win,7,1,"gdop","%12.6f",data,"gdop")
                            _add_value_str(sky_win,8,1,"pdop","%12.6f",data,"pdop")
                            sky_win.refresh()
    
                            sat_win.clear()
                            
                            satellites = data.get("satellites")
                            
                            if satellites is not None:
                                sat_win.box()
                                sat_win.addstr(0,2,"sats")
                                sat_win.addstr(1,1," PRN az el ss")
                                
                                sats = nlargest(12,satellites,lambda sd : sd.get("ss"))
                                
                                for i,sd in enumerate(sats):
                                    
                                    _add_value_str(sat_win,i+2,1,None,"%3d",sd,"PRN")
                                    _add_value_str(sat_win,i+2,5,None,"%3d",sd,"az")
                                    _add_value_str(sat_win,i+2,9,None,"%2d",sd,"el")
                                    _add_value_str(sat_win,i+2,12,None,"%2d",sd,"ss")
                             
                            sat_win.refresh()

                except InterruptedError:
                    log.info("Wait for GPSD response has been interrupted.")


def usage(argv0):
    print("Usage: %s [options...]"%argv0,file=sys.stderr)
    print("  Supported options are:",file=sys.stderr)
    print("    -u <url>     Specify the URL of the GPSD source connection.",file=sys.stderr)
    print("                 Sample URLs are:.",file=sys.stderr)
    print("             tcp://localhost:2947 The default GPSD connection.",file=sys.stderr)
    print("             file:///my/file.gpstrack A GPS Track file.",file=sys.stderr)
    print("             file:///my/file.gpstrack.gz A gzipped GPS Track file.",file=sys.stderr)
    print("             file:///my/file.gpstrack?interval=1 A GPS Track file replayed",file=sys.stderr)
    print("                                                 in 1 second intervals",file=sys.stderr)
    print("             file:///my/file.gpstrack?interval=1?fake_timestamps=true",file=sys.stderr)
    print("                         A GPS Track file replayed in 1 second intervals",file=sys.stderr)
    print("                         and current timestamps",file=sys.stderr)
    print("             file:///my/file.gpstrack?interval=1?start_timestamp=2017-04-19T22:34:57Z",file=sys.stderr)
    print("                         A GPS Track file replayed in 1 second intervals and",file=sys.stderr)
    print("                         timestamps shifted, so that the first TPV has the given time.",file=sys.stderr)
    print("    -t <timeout> Specify the timeout in seconds to wait for JSON packages",file=sys.stderr)
    print("                 The default value is 10.",file=sys.stderr)
    print("                 A comma-separated pair like 30,5 might be given, which",file=sys.stderr)
    print("                 configures an initial timeout before the first TPV record",file=sys.stderr)
    print("                 of 30 seconds and a followup-timeout of 5 seconds.",file=sys.stderr)
    print("    -l <logfile> Write the received JSON packages to a new gpstrack",file=sys.stderr)
    print("                 logfile. If <logfile> ends with '.gz', the resulting",file=sys.stderr)
    print("                 gpstrack file will be gzip compressed.",file=sys.stderr)
    print("    -c           Activate the curses interface.",file=sys.stderr)
    print("    -q           Quiet mode, useful in combination with -l.",file=sys.stderr)
    print("    -i           Print statistics about the given gpstrack.",file=sys.stderr)
    print("    -u <url>     Specify the URL of the GPSD source connection.",file=sys.stderr)
    print("    -s <server_url> Specify an URL to listen on and replicate the GPSD",file=sys.stderr)
    print("                 JSON packages to. If you need a permanent server, install",file=sys.stderr)
    print("                 the gpstrack-runner package.",file=sys.stderr)
    print("    -I <interval> Specify the interval for serving TPV records on the",file=sys.stderr)
    print("                 specified server socket, the default vaule is 1 second.",file=sys.stderr)

def run_client(argv):
    
    url = urlparse("tcp://localhost:2947")
    logfile = None
    quiet = False
    serve = None
    interval = 1
    timeout = 10
    print_infos = False
    initial_timeout=10
    
    try:
        opts, args = getopt.getopt(argv, "u:l:s:I:t:cqi", ["help","url=","logfile=","serve=","interval=","timeout=","curses","quiet","information"])
    except getopt.GetoptError as err:
        # print help information and exit:
        print(err,file=sys.stderr)  # will print something like "option -a not recognized"
        usage(sys.argv[0])
        sys.exit(64)
    
    use_curses = False
    
    for o, a in opts:
        if o in ("--help"):
            usage(sys.argv[0])
            sys.exit(0)
        elif o in ("-u", "--url"):
            url = urlparse(a)
        elif o in ("-l", "--logfile"):
            logfile = a
        elif o in ("-s", "--server"):
            serve = a
        elif o in ("-c", "--curses"):
            use_curses = True
        elif o in ("-q", "--quiet"):
            quiet = True
        elif o in ("-i", "--information"):
            print_infos = True
        elif o in ("-I", "--interval"):
            try:
                interval = float(a)
            except:
                log.error("Wrong argument [%s] for option [%s]"%(a,o))
                sys.exit(64)
        elif o in ("-t", "--timeout"):
            try:
                
                if ',' in a:
                    t12 = a.split(',',1)
                    initial_timeout = float(t12[0])
                    timeout = float(t12[1])
                else:
                    timeout = float(a)
                    initial_timeout = timeout
            except:
                log.error("Wrong argument [%s] for option [%s]"%(a,o))
                sys.exit(64)
        else:
            assert False, "unhandled option"
        
    try:

        signal.signal(signal.SIGINT,sighandler)
        signal.signal(signal.SIGHUP,sighandler)
        signal.signal(signal.SIGTERM,sighandler)
        signal.signal(signal.SIGQUIT,sighandler)

        if serve is not None:
            run_server(serve,url,interval,_register_interruptor)
        else:
                
            if use_curses:
                curses.wrapper(run_gps_client_curses,url,logfile,initial_timeout=initial_timeout,timeout=timeout)
            else:
                run_gps_client(url,logfile,quiet=quiet,initial_timeout=initial_timeout,timeout=timeout,print_infos=print_infos)
    except:
        log.exception("Error running GPSD client")
