'''
Created on 19. Juli 2016

@author: wglas
'''

import socket
import select
import json
import logging
from urllib.parse import urlparse

from clazzes.gpsclient.helpers import watch_command
import gzip

log=logging.getLogger(__name__)

class GPSDClient(object):
    ''' A minimal client for GPSD fulfilling the contract of a context manager.'''

    def __init__(self,export_comments=False):
        self.sock = None
        self.line_reader = None
        self.source = None
        self.export_comments = export_comments

    def connect(self, addr="tcp://localhost"):
        '''
        Connect this client to the given address.
        :param addr: URL to connect to, default value is ``localhost``, where the default
                 port is ``2947``, if no port is specified in the URL. An urllib
                 ParseResult object may also be passed.
        '''
        if self.sock or self.line_reader:
            raise IOError("GPDClient already connected")
        
        if isinstance(addr,str):
            addr = urlparse(addr)
        
        if addr.scheme == "tcp":
            
            host = addr.hostname
            port = addr.port if addr.port else 2947
        
            addrinfos = socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM)
            for addrinfo in addrinfos:
                family, socktype, proto, _canonname, host_port = addrinfo
                try:
                    self.sock = socket.socket(family,socktype,proto)
                    self.sock.connect(host_port)
                    self.sock.setblocking(False)
                    self.line_reader = self.sock.makefile()
                    break
    
                except (OSError, IOError):
                    log.exception('GPSDSocket.connect(%s:%d) failed'%(host_port[0],host_port[1]))
                    self.close()
            
            if self.line_reader is None:
                raise IOError("Cannot connect to GPSD under [%s:%d]."%(host,port))

        elif addr.scheme == "unix":
            self.sock = socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
            self.sock.connect(addr.path)
            self.line_reader = self.sock.makefile()

        elif addr.scheme == "" or addr.scheme == "file":
            
            if addr.path.endswith(".gz"):
                self.line_reader = gzip.open(addr.path,"rt",encoding='utf-8')
            else:
                self.line_reader = open(addr.path,"r",encoding="utf-8")

        else:
            raise ValueError("Invalid scheme [%s] for GPSD URL."%addr.scheme)
        
        self.source = addr.geturl()

    def watch(self, enable=True, devicepath=None):
        '''
        Start or disable JSON watching on the socket.
        :param enable: Whether to enable or disable watching.
        :param devicepath: An optional path for selecting an individual GPS receiver among others.
        '''
        
        if self.line_reader is None:
            raise IOError("Try to send watch command to an unconnected GPSD client.")
        
        if self.sock is not None:
            command = watch_command(enable,devicepath)
    
            self.sock.sendall(command.encode(encoding='utf_8'))
        

    def next(self, timeout=0):
        '''
        Return a parsed dictionary of a response or ``None``, if no data is available within
        the specified timeout in seconds.
        
        The structure of the returned dictionaries is documented at the `GPSD JSON documentation website <http://catb.org/gpsd/gpsd_json.html>`_
        
        :param timeout: A timeout in seconds, if not specified or zero, wait forever for new
                        responses.
        '''
        
        line = None
        
        while line is None:
        
            line = self.line_reader.readline()
            
            while not line.endswith("\n"):
        
                if self.sock is None:
                    # EOF from file
                    if len(line) == 0:
                        return None
                else:    
                    waitin,_waitout,_waiterror = select.select((self.sock,),(),(),timeout)
                    
                    if not waitin:
                        return None
                
                
                piece = self.line_reader.readline()
                
                if len(piece) == 0:
                    if len(line) > 0:
                        break
                    else: 
                        return None
                
                line += piece
        
            # skip empty lines or comments
            if line == '\n' or line == '\r\n' or line[0] == '#':
                if log.getEffectiveLevel() <= logging.DEBUG:
                    log.debug("Ignoring comment [%s]"%line)
                
                if self.export_comments and line[0] == '#':
                    return { 'class':'COMMENT', 'comment': line.strip() }
                
                line = None
            
        try:
            return json.loads(line)
        except:
            log.exception("Error parsing JSON line [%s] from GPSD"%line)
            raise

    def close(self):
        '''
        Close the socket. This object may be reconnected by ``connect()``
        after this function succeeded.
        '''
        if self.line_reader:
            self.line_reader.close()

        if self.sock:
            self.sock.close()
        
        self.line_reader = None
        self.sock = None

    def __str__(self):
        return self.source

    def __enter__(self):
        return self

    def __exit__(self,exc_type,exc_value,traceback):
        self.close()
        return False