'''

Support for JSON-based GPS log files.

'''

import pytz
import socket
import pkg_resources
from clazzes.tools import datetools
from datetime import datetime
from clazzes.gpsclient.helpers import watch_command, DelegatorSupport
import json
import logging
import gzip

log = logging.getLogger(__name__)

class LogWriter(DelegatorSupport):
    '''
    A JSON logfile writer, which acts as a delegator to anothe GPSD client.
    '''
    
    def __init__(self, delegate = None, tz = None, nodename = None):
        '''
        Instantiate a log writer 
        :param delegate: An optional delegate GPS client to read from.
                The log writer acts as a filter for GPSD events,
                if a delegate is specified.         
        :param tz: An optional tzinfo object or timezone name, if timestamps
               should be logged in a timezone different from the operating
               system's timezone.
        :param nodename: The computer's name as being logged in the header
               of the generated log file. If not specified the name as returned
               by ``socket.gethostname()`` is used.
        '''
        
        super().__init__(delegate)
        
        if tz is None:
            tz  = datetools.get_os_timezone()
            
        if isinstance(tz,str):
            self.tz  = pytz.timezone(tz)
        else:
            self.tz = tz
        
        if nodename is None:
            self.nodename = socket.gethostname()
        else:
            self.nodename = nodename

        self.fh = None
        
        self.npackages = 0
        self.ntpvs = 0
        self.last_time = None
        
    def _print_time_line(self):
        print("# time: %s"%datetools.format_datetime(datetime.now(tz=self.tz)),file=self.fh)
    
    def open(self,filename):
        '''
        Open a new log file for writing.
        :param filename: The filename to write to.
        '''

        if self.fh:
            raise IOError("GPS LogWriter already connected")

        if filename.endswith(".gz"):
            self.fh = gzip.open(filename,"wt",encoding='utf-8')
        else:
            self.fh = open(filename,"w",encoding="utf-8")
        
        try:
            version = pkg_resources.get_distribution("clazzes_gpsclient").version
        except pkg_resources.DistributionNotFound:
            version = "svn"
        
        print("#",file=self.fh)
        print("# JSON GPS log file generated by gpsclient-%s"%version,file=self.fh)
        print("# host: %s"%self.nodename,file=self.fh)
        self._print_time_line()
        print("#",file=self.fh)

    def watch(self, enable=True, devicepath=None):
        '''
        Send a watch command to the delegate and log the watch command as a comment.
        :param enable: Whether to enable or disable watching.
        :param devicepath: An optional path for selecting an individual GPS receiver among others.
        '''
        
        super().watch(enable=enable,devicepath=devicepath)

        print("# "+watch_command(enable,devicepath),file=self.fh,end='')

    def _log_packages(self):
        log.info("Wrote [%d] JSON packages with [%d] TPV records among them, last timestamp was [%s]."%(self.npackages,self.ntpvs,self.last_time))

    def write_package(self,obj):
        '''
        Write a dictionary object as a package to the log file alongside with
        a timestamp line. 
        :param obj: A dictionary to log in JSON format.
        '''
        self._print_time_line()
        print(json.dumps(obj,sort_keys=True),file=self.fh,flush=True)
        self.npackages += 1
        
        cls = obj.get("class")
        
        new_timestamp = False
        
        if cls == "TPV":
            
            self.ntpvs += 1
            time = obj.get("time")
            
            if time is not None and self.last_time is None:
                new_timestamp = time
            
            self.last_time = time
            
        if new_timestamp or self.npackages == 1 or self.npackages == 10 or self.npackages == 100 or self.npackages%1000 == 0:
            self._log_packages()

    def next(self, timeout=0):
        '''
        Fetch the next package from the delegate and log a received
        package to the log file.
        :param timeout: The timeout in seconds to wait for a package on the
            the delegate.
        '''
        
        obj = super().next(timeout=timeout)
        
        if obj is not None:
            self.write_package(obj)

        return obj

    def close(self):
        '''
        Close the log file and the delegate and write a trailing line to the
        log file.
        '''
        if self.fh is not None:
            
            self._print_time_line()
            print("# END OF JSON GPS log file.",file=self.fh)

            self._log_packages()
            
            fn = self.fh.name
            
            try:

                log.info("Closing JSON log file [%s]."%fn)
                self.fh.close()
            except:
                log.exception("Error closing written JSON log file [%s]"%fn)
                
            self.fh = None

        super().close()
        self.npackages = 0
        self.ntpvs = 0
        self.last_time = None
