'''
Helper functions
'''
import json
import logging

log = logging.getLogger(__name__)

def watch_command(enable=True, devicepath=None):
    '''
    Generate a JSON watch command to send to the socket.
    :param enable: Whether to enable or disable watching.
    :param devicepath: An optional path for selecting an individual GPS receiver among others.
    '''

    cmd = { "json": True, "enable" : enable }
        
    if devicepath is not None:
        cmd["device"] = devicepath
        
    return '?WATCH='+json.dumps(cmd)+'\n'

class DelegatorSupport:
    
    def __init__(self,delegate):
        '''
        Instantiate a log writer 
        :param delegate: An optional delegate GPS client to read from.
        '''
        self.delegate = delegate

    def connect(self, addr="tcp://localhost"):
        '''
        Connect the delegate to the given URL. If no delegate is configured,
        this method does nothing.
        :param addr: The URL to connect to, supported schemes are ``tcp``, ``unix`` or ``file``.
        '''
        if self.delegate is not None:
            self.delegate.connect(addr)

    def watch(self, enable=True, devicepath=None):
        '''
        Send a watch command to the delegate and log the watch command as a comment.
        :param enable: Whether to enable or disable watching.
        :param devicepath: An optional path for selecting an individual GPS receiver among others.
        '''
        
        if self.delegate is not None:
            self.delegate.watch(enable,devicepath)
            
    def next(self, timeout=0):
        '''
        Fetch the next package from the delegate.
        If no dlegate is specified, ``None`` is returned.
        :param timeout: The timeout in seconds to wait for a package on the
            the delegate.
        '''
        
        if self.delegate is None:
            return None
            
        return self.delegate.next(timeout=timeout)

    def close(self):
        '''
        Close the delegate,
        '''
        
        if self.delegate is not None:
            try:
                self.delegate.close()

            except:
                log.exception("Error closing GPS Client delegate")

    def __enter__(self):
        return self
    
    def __exit__(self,exc_type,exc_value,traceback):
        self.close()
        return False