'''

Server connection handler for replaying GPS JSON log files.

'''

import json
import logging
import os
import select
import socketserver
import time
from urllib.parse import urlparse

import errno
import threading
from clazzes.gpsclient.factory import new_client


log = logging.getLogger(__name__)

class GpsLogRequestHandler(socketserver.BaseRequestHandler):
    
    def _check_interrupt(self):
        with self.server.gps_lock:
            return self.server.gps_interrupted
    
    def _wait_for_watch(self,ca):
        
        watch = None
        
        while watch is None or not watch.endswith(b"\n"):
        
            part = self.request.recv(256)
            
            if not part:
                raise IOError("Client [%s] prematurely closed the connection before receiving a watch command."%ca)
        
            if watch:
                watch += part
            else:
                watch = part
                
            if len(watch) >= 256:
                raise IOError("Client [%s] sent more than 255 bytes as watch command, closing connection."%ca)
        
        watch_str = watch.decode("utf-8").strip()
        
        if not watch_str.startswith("?WATCH="):
            raise IOError("Client [%s] sent invalid watch command [%s], closing connection."%(ca,watch_str))
        
        params = json.loads(watch_str[7:])
        
        if params["enable"] != True:
            raise IOError("Client [%s] watch command [%s] with enable!=true."%(ca,watch_str))
        
        if params["json"] != True:
            raise IOError("Client [%s] watch command [%s] with json!=true."%(ca,watch_str))
    
    def handle(self):
        
        logfile = self.server.gps_logfile
        timeout = self.server.gps_timeout
        ca = self.client_address[0] if len(self.client_address) > 0 else '<local>'
        
        watching = False
        
        with new_client(logfile) as client:
            
            try:
                log.info("Start streaming of log file [%s] to client [%s]."%(logfile.geturl(),ca))

                client.connect(logfile)
            
                client.watch()
                
                last_clockval = None
            
                while not self._check_interrupt():
                    
                    data = client.next(timeout=10)
                    
                    if data is None:
                        log.info("No more data available from log file [%s], closing connection to client [%s]."%(client,ca))
                        break
                    
                    cls = data.get("class")
                    
                    if cls == "DEVICES" and not watching:
                        log.info("Waiting for WATCH command from client [%s] before delivering DEVICES package."%(ca,))
                        self._wait_for_watch(ca)
                        log.info("Successfully received WATCH command from client [%s]."%(ca,))
                        watching = True
                    
                    if cls == "TPV":
                    
                        clockval = time.monotonic() 
                    
                        if last_clockval is not None:
                            elapsed = clockval-last_clockval
                            wait_time = timeout - elapsed
                            
                            if wait_time > 0:
                                
                                waitin,_waitout,_waiterror = select.select((self.request,self.server.gps_input),(),(),wait_time)
                    
                                if waitin:
                                
                                    if self.request in waitin:
                                        probe = self.request.recv(1)
                                        
                                        if probe:
                                            log.warn("Peer sent additional data while streaming from log file [%s], closing connection to client [%s]."%(client,ca))
                                        else:
                                            log.info("Peer closed connection while streaming from log file [%s], closing connection to client [%s]."%(client,ca))
                                    else:
                                        log.info("Closing connection while streaming from log file [%s] to client [%s]."%(client,ca))
                                    break
                                
                                last_clockval += timeout

                                #
                                # fake timestamps again after wait timeout, because
                                # otherwise the timestamp lags behind by the wit time
                                #
                                if hasattr(client,"fake_timestamp"):
                                    client.fake_timestamp(data)
                                
                            elif wait_time < 0:
                                log.warn("Data from log file [%s] is arriving too slow for interval [%s]"%(client,timeout))
                                last_clockval = clockval
                        
                        else:
                            last_clockval = clockval
                    
                    msg = json.dumps(data,sort_keys=True) + "\n"
                    
                    self.request.sendall(msg.encode("utf-8"))

            except:
                log.exception("Caught exception streaming log file [%s], closing connection to client [%s]."%(logfile.geturl(),ca))

def run_server(serve_url,logfile,timeout=1.0,register_hook=None):
    
    if isinstance(serve_url,str):
        serve_url = urlparse(serve_url)
    
    if isinstance(logfile,str):
        logfile = urlparse(logfile)

    if serve_url.scheme == 'tcp':
    
        port = serve_url.port if serve_url.port else 2948
    
        if logfile.scheme == serve_url.scheme:
            
            logport = logfile.port if logfile.port else 2947
            if port == logport:
                raise ValueError("Cannot serve TCP input [%s] on TCP serve URL [%s] with same port."%(logfile.geturl(),serve_url.geturl()))
    
        log.info("Starting TCP/IP server on [%s:%d]"%(serve_url.hostname,port))
    
        server = socketserver.ThreadingTCPServer((serve_url.hostname,port),GpsLogRequestHandler,bind_and_activate=False)
        
    elif serve_url.scheme == 'unix':
    
        if logfile.scheme == serve_url.scheme and logfile.path == serve_url.path:
            raise ValueError("Cannot serve UNIX input [%s] on UNIX serve URL [%s] with same path."%(logfile.geturl(),serve_url.geturl()))

        log.info("Starting UNIX stream server on [%s]"%(serve_url.path))
    
        server = socketserver.ThreadingUnixStreamServer(serve_url.path,GpsLogRequestHandler,bind_and_activate=False)

    else:
        raise ValueError("Unknown scheme [%s] in serve URL [%s]"%(serve_url.scheme,serve_url))
    
    input = None
    output = None
      
    try:
        
        input,output = os.pipe()
        
        server.gps_timeout = timeout
        server.gps_logfile = logfile
        server.gps_interrupted = False
        server.gps_lock = threading.Lock()
        server.gps_input = input
        server.allow_reuse_address = True

        if register_hook is not None:
            
            def interruptor():
                
                # immediately terminate waiting client connections.
                os.write(output,b"x")
                
                # inform client connection handlers about the end
                with server.gps_lock:
                    server.gps_interrupted = True
                
                # wait a bit to close client connections.
                time.sleep(timeout * 0.1)
                
                # see https://stackoverflow.com/questions/17550389
                log.info("Shutting down server on [%s] upon interrupt."%(serve_url.geturl()))
                t = threading.Thread(target = server.shutdown)
                t.start()
            
            register_hook(interruptor)

        server.server_bind()
        server.server_activate()
        server.serve_forever()
    
    except OSError as e:    
        if e.errno == errno.EBADF:
            log.info("Server on [%s] has been stopped by signal."%(serve_url.geturl()))
        else:
            log.exception("Server on [%s] has been stopped by unexpected exception"%(serve_url.geturl()))
            
    finally:
        server.server_close()
        
        if input is not None:
            os.close(input)
            
        if output is not None:
            os.close(output)
