'''
   Loop support for files or directories.
'''
from clazzes.gpsclient.helpers import DelegatorSupport
from clazzes.gpsclient.gpsd import GPSDClient

import os
import stat
import logging
from urllib import parse
from clazzes.tools.lang import str2bool

log = logging.getLogger(__name__)

class Loop(DelegatorSupport):
    
    def __init__(self, endless=True, export_comments=False):
    
        super().__init__(GPSDClient(export_comments = export_comments))
        self.urls = None
        self.current_index = None
        self.endless = endless
    
    def _connect_next(self):
            
        if self.current_index is None:
            self.current_index = 0
        elif self.endless:
            self.current_index = (self.current_index + 1) % len(self.urls)
        else:
            self.current_index += 1
        
        if self.current_index < len(self.urls):

            url = self.urls[self.current_index]          
            log.info("Starting cycle for [%s]"%url)
            super().connect(addr=url)
    
    def connect(self, addr="tcp://localhost"):
        '''
        Connect to a file or directory URL.
        If a directory is given, the directory is searched for files matching the globs
        ``*.gps-log`` or ``*.gps-log`` and the list of the matched fils is iterated
        when calling the next() method.
        :param addr:
        '''
        
        if self.urls is not None:
            raise IOError("Loop already connected")

        if isinstance(addr,str):
            addr = parse.urlparse(addr)

        endless_s = parse.parse_qs(addr.query).get('endless')
        
        if endless_s is not None:
            self.endless = str2bool(endless_s[0])

        if addr.scheme == '' or addr.scheme == "file":
            
            stat_res = os.stat(addr.path)
            
            if stat.S_ISDIR(stat_res.st_mode):
                
                log.info("Scanning directory [%s] for JSON logfiles..."%addr.path)
                
                all_files = os.listdir(path=addr.path)
                
                matching_files = [file for file in all_files if file.endswith('.gpstrack') or file.endswith('.gpstrack.gz')]
                matching_files.sort()
                
                self.urls = []
                
                for file in matching_files:
                    
                    file_url = parse.quote(os.path.join(addr.path,file))
                    log.info("Found JSON log file [%s]."% file_url)
                    self.urls.append(file_url)
                
                log.info("Found [%d] JSON log files in directory [%s]."%(len(self.urls),addr.path))       
        
        if self.urls is None:
            self.urls = [ addr ]
            
        self._connect_next()
        
    def next(self, timeout=0):
        '''
        Return the next data set from the currently open file in the loop
        or turn on to the next file in loop, if the current file has reaches EOF. 
        :param timeout: The maximal number of seconds to wait for a new dataset.
        '''
        
        if self.urls is None:
            raise IOError("Try to get a dataset from an unconnected Loop.")
        
        if self.current_index < len(self.urls):
            
            ret = super().next(timeout=timeout)
            
            if ret is None:
                
                # EOF from individual file -> open next one.
                log.info("Cycle for [%s] finished."%str(self.delegate))

                super().close()
                self._connect_next()
            
                # we may have reached the end of a non-endless loop,
                # so we may return None below.
                if self.current_index < len(self.urls):
                    ret = super().next(timeout=timeout)
                    
            return ret
            
        else:
            return None
    
    def close(self):
        
        super().close()
        self.urls = None
        self.current_index = None
    
    def __str__(self):
        base = str(self.delegate)
        return "%s%sendless=%s"%(base,'&' if '?' in base else '?',"true" if self.endless else "false")