'''

  A throttle, which outputs a client's TPV records in a fixed
  interval.

'''
from clazzes.gpsclient.helpers import DelegatorSupport
import time
import logging
from urllib.parse import urlparse, parse_qs

log = logging.getLogger(__name__)

class Throttle(DelegatorSupport):
    
    def __init__(self,delegate,interval = 1):
        '''
        :param delegate: The delegate to fetch GPSD JSON packages from.
        :param interval: The interval to output the records in seconds.
        '''
    
        super().__init__(delegate)
        self.last_clockval = None
        self.interval = interval
    
    def connect(self, addr="tcp://localhost?interval=1"):
        
        if isinstance(addr,str):
            addr = urlparse(addr)

        interval_s = parse_qs(addr.query).get('interval')
        
        if interval_s is not None:
            self.interval = float(interval_s[0])
            
        super().connect(addr=addr)
    
    def next(self, timeout=0):
        '''
        Fetch the next package from the delegate and output the received
        package not after the configured interval after last TPV record received,
        if the newly arriving package is a TPV package, too.
        
        :param timeout: The timeout in seconds to wait for a package on the
            the delegate. This timeout has to be larger than the interval specified
            in the constructor.
        '''
        obj = super().next(timeout=timeout)
        
        if obj is not None:
            
            if obj.get("class") == "TPV":
                    
                clockval = time.monotonic()
                
                if self.last_clockval is not None:
                    elapsed = clockval-self.last_clockval
                    wait_time = self.interval - elapsed
                            
                    if wait_time > 0:
                        
                        time.sleep(wait_time)
                                    
                        self.last_clockval += self.interval
                                
                    elif wait_time < 0:
                        log.warn("Data from source [%s] is arriving too slow for interval [%s]"%(str(self.delegate),self.interval))
                        self.last_clockval = clockval
                        
                else:
                    self.last_clockval = clockval

        return obj

    def __str__(self):
        
        base = str(self.delegate)
        return "%s%sinterval=%s"%(base,'&' if '?' in base else '?',self.interval)
